/* Copyright (C) 2008 by Attila Tóth <atoth@atoth.sote.hu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* otherwise compiler complains of strcasestr making integer from a pointer
 * without a cast. It's a non-standard extension. It's easier this way.
 * However stsstr doesn't need this. See:
 * http://www.mail-archive.com/freeciv-dev@gna.org/msg03863.html
 * http://stackoverflow.com/questions/9935642/how-do-i-use-strcasestr
 */
#define _GNU_SOURCE

#include <string.h>		/* memset(), strcmp(), strncmp() */
#include <stdlib.h>		/* strtoul(), malloc(), free(), exit() */
#include <stdio.h>		/* printf(), fprintf() */
#include <sysexits.h>           /* EX_* */
#include <syslog.h>             /* openlog(), syslog(), closelog() */
#include <sys/types.h>		/* getaddrinfo(), freeaddrinfo(), fork(), open() */
#include <netdb.h>              /* getaddrinfo(), freeaddrinfo() */
#include <netinet/in.h>         /* struct sockaddr_in */
#include <ctype.h>              /* isspace() */
#include <unistd.h>		/* fork(), setsid(), close(), gethostname(), getdomainname() */
#include <sys/stat.h>		/* open() */
#include <fcntl.h>		/* open() */
#include <errno.h>		/* errno */
#include <limits.h>             /* UINT_MAX */

#include <libmilter/mfapi.h>

#define MIN(a, b) (a < b ? a : b)


/* taken from: http://stackoverflow.com/questions/779875/what-is-the-function-to-replace-string-in-c */
char *replace(char const * const original, char const * const pattern, char const * const replacement);


/*
This is part of the libb64 project, and has been placed in the public domain.
For details, see http://sourceforge.net/projects/libb64
*/

typedef enum
{
	step_a, step_b, step_c, step_d
} base64_decodestep;

typedef struct
{
	base64_decodestep step;
	char plainchar;
} base64_decodestate;

void base64_init_decodestate(base64_decodestate* state_in);

int base64_decode_value(char value_in);

int base64_decode_block(const char* code_in, const int length_in, char* plaintext_out, base64_decodestate* state_in);
/* libb64 end */


struct mlfiPriv
{
	char *mailing_list;
	char *subject;
	char *header_from;
	char *header_to;
	char *header_cc;
/*	char *mail_hostp;
	char *mail_addrp;
	char *rcpt_hostp; */
	char *rcpt_addrp;
	char *return_path;
	unsigned char listness;
	unsigned char subjectness;
	unsigned char clue;
	unsigned char have_from;
	unsigned char have_to;
	unsigned char have_cc;
	unsigned char have_subject;
	unsigned char dupl_subject;
	unsigned char subject_idx;
	unsigned char local_mail;
	unsigned char clamavness;
	unsigned char spamness;
	unsigned char returnpath;
};

char **list_array = NULL;

int listno = 0;
/* The following are all options set on the command line */
unsigned char debug = 0;
char *sock_addr = NULL;
char *conf_file = NULL;
char *host_name = NULL;
char *domain_name = NULL;
char *fqdn = NULL;

/*
 *   Initialization for the message
 */
sfsistat mlfi_connect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *hostaddr)
{
	struct mlfiPriv *priv;

	priv = malloc(sizeof *priv);
	if (priv == NULL)
	{
		syslog(LOG_DEBUG, "Malloc for milter priv struct failed (location: 00).\n");
		return SMFIS_TEMPFAIL;
	}
	memset(priv, '\0', sizeof *priv);

	priv->mailing_list = NULL;
	priv->subject = NULL;
/*	priv->mail_hostp = NULL;
	priv->mail_addrp = NULL;
	priv->rcpt_hostp = NULL; */
	priv->rcpt_addrp = NULL;
	priv->header_from = NULL;
	priv->header_to = NULL;
	priv->header_cc = NULL;
	priv->return_path = NULL;
	priv->listness = 0;
/*	priv->subjectness = 1; */
	priv->subjectness = 0;
	priv->clue = 0;
	priv->have_from = 0;
	priv->have_to = 0;
	priv->have_cc = 0;
	priv->have_subject = 0;
	priv->dupl_subject = 0;
	priv->subject_idx = 0;
	priv->local_mail = 0;
	priv->clamavness = 0;
	priv->spamness = 0;
	priv->returnpath = 0;

	smfi_setpriv(ctx, priv);

	return SMFIS_CONTINUE;
}

/* 
 *   Check for list addresses and set the necessary fields upon match
 */
/* sfsistat mlfi_envfrom(SMFICTX *ctx, char **argv)
{
	struct mlfiPriv *priv = smfi_getpriv(ctx);
	char *mail_host = NULL;
	char *mail_addr = NULL;
	int i;
	size_t len;

	len = (strlen(smfi_getsymval(ctx, "{mail_addr}")) + 1);
	if (len == 0)
	{
		syslog(LOG_DEBUG, "Realloc for mail_addr: len = 0 (location: 01a).\n");
		return SMFIS_TEMPFAIL;
	}
	snprintf(mail_addr, len, "%s", smfi_getsymval(ctx, "{mail_addr}"));
	priv->mail_addrp = strdup(mail_addr);
	mail_addr = realloc(mail_addr, len * sizeof(char));
	if (mail_addr == NULL)
	{
		syslog(LOG_DEBUG, "Malloc for mail_addr failed (location: 01b).\n");
		return SMFIS_TEMPFAIL;
	}
	snprintf(mail_addr, len, "%s", smfi_getsymval(ctx, "{mail_addr}"));
	priv->mail_addrp = strdup(mail_addr);

	free(mail_addr);

	len = (strlen(smfi_getsymval(ctx, "{mail_host}")) + 1);
	if (len > 1)
	{
		mail_host = realloc(mail_host, len * sizeof(char));
		if (mail_host == NULL)
		{
			syslog(LOG_DEBUG, "Malloc for mail_host failed (location: 02).\n");
			return SMFIS_TEMPFAIL;
		}
		snprintf(mail_host, len, "%s", smfi_getsymval(ctx, "{mail_host}"));
		priv->mail_hostp = strdup(mail_host);

		free(mail_host);
	}

	if (debug == 1)
	{
		syslog(LOG_DEBUG, "{mail_addr}: %s.", smfi_getsymval(ctx, "{mail_addr}"));
		syslog(LOG_DEBUG, "{mail_host}: %s.", smfi_getsymval(ctx, "{mail_host}"));
	}

	return SMFIS_CONTINUE;
} */

sfsistat mlfi_envrcpt(SMFICTX *ctx, char **argv) {
	struct mlfiPriv *priv = smfi_getpriv(ctx);
	char *rcpt_addr = NULL;
/*	char *rcpt_host = NULL; */
	int i;
	size_t len;

	len = (strlen(smfi_getsymval(ctx, "{rcpt_addr}")) + 1);
	if (len == 0)
	{
		syslog(LOG_DEBUG, "Realloc for rcpt_addr: len = 0 (location: 03a).\n");
		return SMFIS_TEMPFAIL;
	}
	rcpt_addr = realloc(rcpt_addr, len * sizeof(char));
	if (rcpt_addr == NULL)
	{
		syslog(LOG_DEBUG, "Malloc for rcpt_addr failed (location: 03b).\n");
		return SMFIS_TEMPFAIL;
	}
	snprintf(rcpt_addr, len, "%s", smfi_getsymval(ctx, "{rcpt_addr}"));
	priv->rcpt_addrp = strdup(rcpt_addr);

	for (i = 0; i < listno; i++)
	{
		if (priv->clue == 0)
		{
			if (strcmp(rcpt_addr, list_array[i]) == 0)
			{
				priv->clue = 1;

				if (debug == 1)
					syslog(LOG_DEBUG, "List-Milter: list_array[%d]: %s, listness: %d, clue: %d.\n", i, list_array[i], priv->listness, priv->clue);
			}
		}
	}

	free(rcpt_addr);

/*	len = (strlen(smfi_getsymval(ctx, "{rcpt_host}")) + 1);
	if (len > 1)
	{
		rcpt_host = realloc(rcpt_host, len * sizeof(char));
		if (rcpt_host == NULL)
		{
			syslog(LOG_DEBUG, "Malloc for rcpt_host failed (location: 04).\n");
			return SMFIS_TEMPFAIL;
		}
		snprintf(rcpt_host, len, "%s", smfi_getsymval(ctx, "{rcpt_host}"));
		priv->rcpt_hostp = strdup(rcpt_host);

		free(rcpt_host);
	} */

	if (debug == 1)
	{
		syslog(LOG_DEBUG, "List-Milter: listness: %d, subjectness: %d, clue: %d.\n", priv->listness, priv->subjectness, priv->clue);
		syslog(LOG_DEBUG, "{rcpt_addr}: %s, rcpt_addrp: %s.\n", smfi_getsymval(ctx, "{rcpt_addr}"), priv->rcpt_addrp);
/*		syslog(LOG_DEBUG, "{rcpt_host}: %s, rcpt_hostp: %s.\n", smfi_getsymval(ctx, "{rcpt_host}"), priv->rcpt_hostp); */
	}

	return SMFIS_CONTINUE;
}

/*
 *   Catching list message
 */
sfsistat mlfi_header(SMFICTX *ctx, char *headerf, char *headerv)
{
	struct mlfiPriv *priv = smfi_getpriv(ctx);
	char *listname_buffer = NULL;
	char *subject_b64dec = NULL;
	char *subject_b64enc = NULL;
/*	char *subject_replaced = NULL; */
	int i, len2, len3;
/*	int len1; -Wall unused variable */
	size_t len;
	base64_decodestate state;

/*
 * Squirrelmail PHP Base64Encode Header function started to play a game, that in case of long lines it inserted
 * a leading newline. That caused subject tag detection to fail. Therefore as a workaround we skip leading newlines.
 */
	if (strcasecmp("Subject", headerf) == 0)
	{
		if (headerv[0] == '\n')
		{
			headerv++;
			headerv++;
			syslog(LOG_DEBUG, "Skipped the first two characters of the subject, because of a leading newline.\n");
		}
	}

	if (debug == 1)
		syslog(LOG_DEBUG, "List-Milter: headerf: %s; headerv: %s; listno: %d.", headerf, headerv, listno);

	if (strcasecmp("To", headerf) == 0)
	{
		if (priv->have_to == 1)
		{
			syslog(LOG_DEBUG, "Second instance of To header using different case combination ignored.\n");
			priv->spamness = 1;
		}
		else if (priv->have_to == 0)
		{
			if (priv->clue == 1)
			{
				for (i = 0; i < listno; i++)
				{
					if (priv->listness == 0)
					{
						len = (strlen(list_array[i]) + strlen(fqdn) + 2);
						if (len == 0)
						{
							syslog(LOG_DEBUG, "Realloc for listname_buffer: len = 0 (location: 05a).\n");
							return SMFIS_TEMPFAIL;
						}
						listname_buffer = (char*) realloc(listname_buffer, len * sizeof(char));
						if (listname_buffer == NULL)
						{
							syslog(LOG_DEBUG, "Malloc for listname_buffer failed (location: 05b).\n");
							return SMFIS_TEMPFAIL;
						}
						snprintf(listname_buffer, len, "%s@%s", list_array[i], fqdn);

						if (debug == 1)
							syslog(LOG_DEBUG, "List-Milter: headerv: %s listname_buffer: %s.", headerv, listname_buffer);

						if (strstr(headerv, listname_buffer) != NULL)
						{
							priv->mailing_list = list_array[i];
							priv->listness = 1;
						}
					}
				}

				free(listname_buffer);
			}
			priv->header_to = strdup(headerv);
			priv->have_to = 1;
		}
		else
		{
			syslog(LOG_DEBUG, "WTF have_to?\n");
		}
	}

	if (strcasecmp("Cc", headerf) == 0)
	{
		if (priv->have_cc == 1)
		{
			syslog(LOG_DEBUG, "Second instance of Cc header using different case combination ignored.\n");
			priv->spamness = 1;
		}
		else if (priv->have_cc == 0)
		{
			if (priv->clue == 1)
			{
				for (i = 0; i < listno; i++)
				{
					if (priv->listness == 0)
					{
						len = (strlen(list_array[i]) + strlen(fqdn) + 2);
						if (len == 0)
						{
							syslog(LOG_DEBUG, "Realloc for listname_buffer: len = 0 (location: 05v).\n");
							return SMFIS_TEMPFAIL;
						}
						listname_buffer = (char*) realloc(listname_buffer, len * sizeof(char));
						if (listname_buffer == NULL)
						{
							syslog(LOG_DEBUG, "Malloc for listname_buffer failed (location: 05w).\n");
							return SMFIS_TEMPFAIL;
						}
						snprintf(listname_buffer, len, "%s@%s", list_array[i], fqdn);

						if (debug == 1)
							syslog(LOG_DEBUG, "List-Milter: headerv: %s listname_buffer: %s.", headerv, listname_buffer);

						if (strstr(headerv, listname_buffer) != NULL)
						{
							priv->mailing_list = list_array[i];
							priv->listness = 1;
						}
					}
				}

				free(listname_buffer);
			}
			priv->header_cc = strdup(headerv);
			priv->have_cc = 1;
		}
		else
		{
			syslog(LOG_DEBUG, "WTF have_cc?\n");
		}
	}

	if (strcasecmp("From", headerf) == 0)
	{
		if (priv->have_from == 1)
		{
			syslog(LOG_DEBUG, "Second instance of From header using different case combination ignored.\n");
			priv->spamness = 1;
		}
		else if (priv->have_from == 0)
		{
			priv->header_from = strdup(headerv);
			priv->have_from = 1;
		}
		else
		{
			syslog(LOG_DEBUG, "WTF have_from?\n");
		}
	}

	if (strcasecmp("Subject", headerf) == 0)
	{
		if (priv->have_subject == 1)
		{
			syslog(LOG_DEBUG, "Second instance of Subject header using different case combination ignored.\n");
			priv->spamness = 1;
			priv->dupl_subject = 1;
			if (headerv == NULL)
			{
				priv->subject_idx = 2;
			}
			else if (strstr(headerv, "SZPEM") != NULL)
			{
				priv->subject_idx = 1;
			}
		}
		else if (priv->have_subject == 0)
		{
			if (priv->clue == 1)
			{
				if (strstr(headerv, "[Atoth: ") != NULL)
				{
					priv->subjectness = 1;
				}
				else if (strstr(headerv, "=5BAtoth=3A_") != NULL)
				{
					priv->subjectness = 1;
				}
				else
				{
					subject_b64enc = strcasestr(headerv, "?B?");
					if (subject_b64enc != NULL)
					{
/*						len1 = (strlen(headerv));
						if (len1 == 0)
						{
							syslog(LOG_DEBUG, "Realloc for subject_replaced: len1 = 0 (location: 06a).\n");
							return SMFIS_TEMPFAIL;
						}
						subject_replaced = (char*) realloc(subject_replaced, len1 * sizeof(char));
						if (subject_replaced == NULL)
						{
							syslog(LOG_DEBUG, "Malloc for subject_replaced failed (location: 06b).\n");
							return SMFIS_TEMPFAIL;
						}
						subject_replaced = replace(headerv, "?=\n =?utf-8?B?", ""); */
						len2 = (strlen(subject_b64enc));
						if (len2 == 0)
						{
							syslog(LOG_DEBUG, "Realloc for subject_replaced: len2 = 0 (location: 06c).\n");
							return SMFIS_TEMPFAIL;
						}
						subject_b64dec = (char*) realloc(subject_b64dec, len2 * sizeof(char));
						if (subject_b64dec == NULL)
						{
							syslog(LOG_DEBUG, "Malloc for subject_b64dec failed (location: 06d).\n");
							return SMFIS_TEMPFAIL;
						}
						base64_init_decodestate(&state);
						len3 = base64_decode_block(subject_b64enc + 3, len2, subject_b64dec, &state);
						if (len2 < len3)
						{
							syslog(LOG_DEBUG, "Base64 encoded text is shorter than decoded!? (location: 07).\n");
							return SMFIS_TEMPFAIL;
						}
						syslog(LOG_DEBUG, "List-Milter: Base64 decoded subject string: %s.", subject_b64dec);
						if (strstr(subject_b64dec, "[Atoth: ") != NULL)
						{
							priv->subjectness = 1;
						}
/*						else
						{
							priv->subjectness = 0;
						} */
						if (debug == 1)
							syslog(LOG_DEBUG, "List-Milter: subject_b64enc: %s.", subject_b64enc);
/*							syslog(LOG_DEBUG, "List-Milter: subject_replaced: %s.", subject_replaced); */
						free(subject_b64dec);
/*						free(subject_replaced); */
					}
					else
					{
						syslog(LOG_DEBUG, "Base64 WTF.\n");
/*						priv->subjectness = 0; */
					}
				}
			}
			else if (priv->clamavness == 1)
			{
				if (strstr(headerv, "SZPEM") != NULL)
				{
					priv->clamavness = 0;
				}
				else
				{
					subject_b64enc = strcasestr(headerv, "?B?");
					if (subject_b64enc != NULL)
					{
						len2 = (strlen(subject_b64enc));
						if (len2 == 0)
						{
							syslog(LOG_DEBUG, "Realloc for subject_b64dec: len2 =0 (location: 08a).\n");
							return SMFIS_TEMPFAIL;
						}
						subject_b64dec = (char*) realloc(subject_b64dec, len2 * sizeof(char));
						if (subject_b64dec == NULL)
						{
							syslog(LOG_DEBUG, "Malloc for subject_b64dec failed (location: 08b).\n");
							return SMFIS_TEMPFAIL;
						}
						len3 = base64_decode_block(subject_b64enc + 3, len2, subject_b64dec, &state);
						if (len2 < len3)
						{
							syslog(LOG_DEBUG, "Base64 encoded text is shorter than decoded!? (location: 09).\n");
							return SMFIS_TEMPFAIL;
						}

/*						if (strstr(subject_b64dec, "SZPEM") != NULL)
						{
							priv->clamavness = 0;
						} */
						free(subject_b64dec);
					}
				}
			}

			priv->subject = strdup(headerv);
			priv->have_subject = 1;

			if (debug == 1)
				syslog(LOG_DEBUG, "List-Milter: subject_b64enc: %s.", subject_b64enc);

		}
		else
		{
			syslog(LOG_DEBUG, "WTF have_subject?\n");
		}
	}

	if (strcasecmp("Message-ID", headerf) == 0)
	{
		if (strstr(headerv, "@atoth.") != NULL)
		{
			priv->local_mail = 1;
		}
	}

	if ((strcasecmp("X-Cron-Env", headerf) == 0) && (priv->local_mail == 0))
	{
		priv->local_mail = 1;
	}

	if (strcmp("X-Virus-Status", headerf) == 0)
	{
		if (strstr(headerv, "Infected") != NULL)
		{
			priv->clamavness = 1;
		}
	}

	if (strcmp("X-Spam-Status", headerf) == 0)
	{
		if (strstr(headerv, "Yes") != NULL)
		{
			priv->clamavness = 0;
			priv->spamness = 1;
		}
	}

	if (strcmp("X-Spam-Flag", headerf) == 0)
	{
		if (strstr(headerv, "YES") != NULL)
		{
			priv->clamavness = 0;
			priv->spamness = 1;
		}
	}

	if (strcmp("X-USN-MailScanner-SpamCheck", headerf) == 0)
	{
		if (strstr(headerv, "not spam") == NULL)
		{
			priv->spamness = 1;
		}
	}

	if (strcasecmp("Return-Path", headerf) == 0)
	{
		if (priv->returnpath == 1)
		{
			syslog(LOG_DEBUG, "Second instance of Return-Path header using different case combination ignored.\n");
			priv->spamness = 1;
		}
		else if (priv->returnpath == 0)
		{
			priv->return_path = strdup(headerv);
			priv->returnpath = 1;
		}
		else
		{
			syslog(LOG_DEBUG, "WTF returnpath?\n");
		}
	}

	return SMFIS_CONTINUE;
}

/*
 *   Add header_content to the message, and add list headers if it is a list
 */
sfsistat mlfi_eom(SMFICTX *ctx)
{
	struct mlfiPriv *priv = smfi_getpriv(ctx);
	char *list_address_buf = NULL;
	char *list_header_buf = NULL;
	char *list_request_buf = NULL;
	size_t len, len2;

	if (debug == 1)
	{
		syslog(LOG_DEBUG, "List-Milter: listness: %d, spamness: %d, clamavness: %d, subjectness: %d.\n", priv->listness, priv->spamness, priv->clamavness, priv->subjectness);
		syslog(LOG_DEBUG, "List-Milter: local_mail: %d, returnpath: %d, have_from: %d, have_to: %d, have_subject: %d.\n", priv->local_mail, priv->returnpath, priv->have_from, priv->have_to, priv->have_subject);
	}

	/* Add header_contents to the message */
	if (priv->spamness == 1)
	{
		if (priv->listness == 0)
		{
			smfi_addheader(ctx, "X-List-Milter", "szpem mail");
			if (debug == 1)
				syslog(LOG_DEBUG, "List-Milter: added header: szpem mail.");
		}
		else if (priv->listness == 1)
		{
			smfi_addheader(ctx, "X-List-Milter", "szpem list mail");
			if (debug == 1)
				syslog(LOG_DEBUG, "List-Milter: added header: szpem list mail - WTF!");
		}
		else
		{
			syslog(LOG_DEBUG, "List-Milter: WTF listness?");
		}
	}
	else if (priv->listness == 1)
	{
		smfi_addheader(ctx, "Precedence", "list");

		len = (strlen(priv->mailing_list) + strlen(fqdn) + 2);
		if (len == 0)
		{
			syslog(LOG_DEBUG, "Realloc for list_address_buf: len = 0 (location: 10a).\n");
			return SMFIS_TEMPFAIL;
		}
		list_address_buf = (char*) realloc(list_address_buf, len * sizeof(char));
		if (list_address_buf == NULL)
		{
			syslog(LOG_DEBUG, "Malloc for list_address_buf failed (location: 10b).\n");
			return SMFIS_TEMPFAIL;
		}
		snprintf(list_address_buf, len, "%s@%s", priv->mailing_list, fqdn);
		smfi_chgheader(ctx, "Reply-To", 1, list_address_buf);
		smfi_addheader(ctx, "Reply-To-Mailing-List", list_address_buf);
		smfi_addheader(ctx, "X-Loop", list_address_buf);
		smfi_addheader(ctx, "X-BeenThere", list_address_buf);
		smfi_addheader(ctx, "X-Original-To", list_address_buf);
		smfi_addheader(ctx, "X-Delivered-To", list_address_buf);
		smfi_addheader(ctx, "Delivered-To", list_address_buf);

		len = (strlen(list_address_buf) + 3);
		if (len == 0)
		{
			syslog(LOG_DEBUG, "Realloc for list_header_buf: len = 0 (location: 11a).\n");
			return SMFIS_TEMPFAIL;
		}
		list_header_buf = (char*) realloc(list_header_buf, len * sizeof(char));
		if (list_header_buf == NULL)
		{
			syslog(LOG_DEBUG, "Malloc for list_header_buf failed (location: 11b).\n");
			return SMFIS_TEMPFAIL;
		}
		snprintf(list_header_buf, len, "<%s.%s>", priv->mailing_list, fqdn);
		smfi_addheader(ctx, "List-Id", list_header_buf);
		smfi_addheader(ctx, "X-List-ID", list_header_buf);

		if ((priv->returnpath == 1) && (priv->return_path != NULL))
		{
			smfi_addheader(ctx, "Old-Return-Path", priv->return_path);
		}

		len = (strlen(list_address_buf) + 3);
		if (len == 0)
		{
			syslog(LOG_DEBUG, "Realloc for list_header_buf: len = 0 (location: 11c).\n");
			return SMFIS_TEMPFAIL;
		}
		list_header_buf = (char*) realloc(list_header_buf, len * sizeof(char));
		if (list_header_buf == NULL)
		{
			syslog(LOG_DEBUG, "Malloc for list_header_buf failed (location: 11d).\n");
			return SMFIS_TEMPFAIL;
		}
		snprintf(list_header_buf, len, "<%s@%s>", priv->mailing_list, fqdn);
		smfi_addheader(ctx, "Mailing-List", list_header_buf);
		smfi_addheader(ctx, "X-Mailing-List", list_header_buf);

		len = (strlen(list_address_buf) + 11);
		if (len == 0)
		{
			syslog(LOG_DEBUG, "Realloc for list_header_buf: len = 0 (location: 12a).\n");
			return SMFIS_TEMPFAIL;
		}
		list_header_buf = (char*) realloc(list_header_buf, len * sizeof(char));
		if (list_header_buf == NULL)
		{
			syslog(LOG_DEBUG, "Malloc for list_header_buf failed (location: 12b).\n");
			return SMFIS_TEMPFAIL;
		}
		snprintf(list_header_buf, len, "<%s-request@%s>", priv->mailing_list, fqdn);
		smfi_chgheader(ctx, "Return-Path", 1, list_header_buf);
		smfi_addheader(ctx, "Resent-Sender", list_address_buf);

		len = (strlen(list_address_buf) + 10);
		if (len == 0)
		{
			syslog(LOG_DEBUG, "Realloc for list_header_buf: len = 0 (location: 13a).\n");
			return SMFIS_TEMPFAIL;
		}
		list_header_buf = (char*) realloc(list_header_buf, len * sizeof(char));
		if (list_header_buf == NULL)
		{
			syslog(LOG_DEBUG, "Malloc for list_header_buf failed (location: 13b).\n");
			return SMFIS_TEMPFAIL;
		}
		snprintf(list_header_buf, len, "<mailto:%s>", list_address_buf);
		smfi_addheader(ctx, "List-Post", list_header_buf);

		len = (strlen(list_address_buf) + 26);
		if (len == 0)
		{
			syslog(LOG_DEBUG, "Realloc for list_request_buf: len = 0 (location: 14a).\n");
			return SMFIS_TEMPFAIL;
		}
		list_request_buf = (char*) realloc(list_request_buf, len * sizeof(char));
		if (list_request_buf == NULL)
		{
			syslog(LOG_DEBUG, "Malloc for list_request_buf failed (location: 14b).\n");
			return SMFIS_TEMPFAIL;
		}
		snprintf(list_request_buf, len, "<mailto:%s-request@%s?subject=", priv->mailing_list, fqdn);

		len = (strlen(list_request_buf) + 6);
		if (len == 0)
		{
			syslog(LOG_DEBUG, "Realloc for list_header_buf: len = 0 (location: 15a).\n");
			return SMFIS_TEMPFAIL;
		}
		list_header_buf = (char*) realloc(list_header_buf, len * sizeof(char));
		if (list_header_buf == NULL)
		{
			syslog(LOG_DEBUG, "Malloc for list_header_buf failed (location: 15b).\n");
			return SMFIS_TEMPFAIL;
		}
		snprintf(list_header_buf, len, "%shelp>", list_request_buf);
		smfi_addheader(ctx, "List-Help", list_header_buf);

		len = (strlen(list_request_buf) + 11);
		if (len == 0)
		{
			syslog(LOG_DEBUG, "Realloc for list_header_buf: len = 0 (location: 16a).\n");
			return SMFIS_TEMPFAIL;
		}
		list_header_buf = (char*) realloc(list_header_buf, len * sizeof(char));
		if (list_header_buf == NULL)
		{
			syslog(LOG_DEBUG, "Malloc for list_header_buf failed (location: 16b).\n");
			return SMFIS_TEMPFAIL;
		}
		snprintf(list_header_buf, len, "%ssubscribe>", list_request_buf);
		smfi_addheader(ctx, "List-Subscribe", list_header_buf);

		len = (strlen(list_request_buf) + 13);
		if (len == 0)
		{
			syslog(LOG_DEBUG, "Realloc for list_header_buf: len = 0 (location: 17a).\n");
			return SMFIS_TEMPFAIL;
		}
		list_header_buf = (char*) realloc(list_header_buf, len * sizeof(char));
		if (list_header_buf == NULL)
		{
			syslog(LOG_DEBUG, "Malloc for list_header_buf failed (location: 17b).\n");
			return SMFIS_TEMPFAIL;
		}
		snprintf(list_header_buf, len, "%sunsubscribe>", list_request_buf);
		smfi_addheader(ctx, "List-Unsubscribe", list_header_buf);

		len = (strlen(list_address_buf) + 9);
		if (len == 0)
		{
			syslog(LOG_DEBUG, "Realloc for list_header_buf: len = 0 (location: 18a).\n");
			return SMFIS_TEMPFAIL;
		}
		list_header_buf = (char*) realloc(list_header_buf, len * sizeof(char));
		if (list_header_buf == NULL)
		{
			syslog(LOG_DEBUG, "Malloc for list_header_buf failed (location: 18b).\n");
			return SMFIS_TEMPFAIL;
		}
		snprintf(list_header_buf, len, "<owner-%s@%s>", priv->mailing_list, fqdn);
		smfi_addheader(ctx, "Errors-To", list_header_buf);

		if (priv->subjectness == 0)
		{
/*			if (priv->subject == NULL)
			{
				syslog(LOG_DEBUG, "priv->subjectness = 0, but priv->subject is NULL! (location: 19)\n");
				return SMFIS_TEMPFAIL;
			} */

			if (priv->have_subject == 1)
			{
				len2 = (strlen(priv->mailing_list) + strlen(priv->subject) + 11);
				len = MIN(len2, 2047);
				if (len == 0)
				{
					syslog(LOG_DEBUG, "Realloc for list_header_buf: len = 0 (location: 20a).\n");
					return SMFIS_TEMPFAIL;
				}
				list_header_buf = (char*) realloc(list_header_buf, len * sizeof(char));
				if (list_header_buf == NULL)
				{
					syslog(LOG_DEBUG, "Malloc for list_header_buf failed (location: 20b).\n");
					return SMFIS_TEMPFAIL;
				}
				snprintf(list_header_buf, len, "[Atoth: %s] %s", priv->mailing_list, priv->subject);
				smfi_chgheader(ctx, "Subject", 1, list_header_buf);
			}
			else
			{
				len = (strlen(priv->mailing_list) + 10);
				if (len == 0)
				{
					syslog(LOG_DEBUG, "Realloc for list_header_buf: len = 0 (location: 23a).\n");
					return SMFIS_TEMPFAIL;
				}
				snprintf(list_header_buf, len, "[Atoth: %s]", priv->mailing_list);
				list_header_buf = (char*) realloc(list_header_buf, len * sizeof(char));
				if (list_header_buf == NULL)
				{
					syslog(LOG_DEBUG, "Malloc for list_header_buf failed (location: 23b).\n");
					return SMFIS_TEMPFAIL;
				}
				snprintf(list_header_buf, len, "[Atoth: %s]", priv->mailing_list);
				smfi_addheader(ctx, "Subject", list_header_buf);
			}

			if (debug == 1)
				syslog(LOG_DEBUG, "List-Milter: changed subject to: %s.", list_header_buf);
			priv->subjectness = 1;
		}

		len = (strlen(priv->mailing_list) + 30);
		if (len == 0)
		{
			syslog(LOG_DEBUG, "Realloc for list_header_buf: len = 0 (location: 21a).\n");
			return SMFIS_TEMPFAIL;
		}
		list_header_buf = realloc(list_header_buf, len * sizeof(char));
		if (list_header_buf == NULL)
		{
			syslog(LOG_DEBUG, "Malloc for list_header_buf failed (location: 21b).\n");
			return SMFIS_TEMPFAIL;
		}
		snprintf(list_header_buf, len, "%s mailing list headers handled", priv->mailing_list);
		smfi_addheader(ctx, "X-List-Milter", list_header_buf);
		if (debug == 1)
		{
			syslog(LOG_DEBUG, "List-Milter: added header: %s.", list_header_buf);
			syslog(LOG_DEBUG, "List-Milter - Prepended headers to message for mailing list: %s.", priv->mailing_list);
		}

		free(list_address_buf);
		free(list_header_buf);
		free(list_request_buf);
	}
	else if (((priv->have_to == 1) && (priv->local_mail == 0) && (strstr(priv->header_to, "@atoth.") != NULL)) &&
		 (((priv->have_from == 1) && (priv->have_to == 1) && ((strstr(priv->header_from, priv->header_to) != NULL) ||
		   (strstr(priv->header_to, priv->header_from) != NULL))) ||
		  ((priv->returnpath == 1) && (priv->have_to == 1) && (strstr(priv->header_to, priv->return_path) != NULL))))
	{
		if (priv->have_subject == 1)
		{
			len2 = (strlen(priv->subject) + 14);
			len = MIN(len2, 2047);
			if (len == 0)
			{
				syslog(LOG_DEBUG, "Realloc for list_header_buf: len = 0 (location: 22a).\n");
				return SMFIS_TEMPFAIL;
			}
			list_header_buf = (char*) realloc(list_header_buf, len * sizeof(char));
			if (list_header_buf == NULL)
			{
				syslog(LOG_DEBUG, "Malloc for list_header_buf failed (location: 22b).\n");
				return SMFIS_TEMPFAIL;
			}
			snprintf(list_header_buf, len, "!Self SZPEM! %s", priv->subject);
			smfi_chgheader(ctx, "Subject", 1, list_header_buf);
			if (debug == 1)
				syslog(LOG_DEBUG, "List-Milter: changed subject to: %s.", list_header_buf);
		}
		else
		{
			list_header_buf = (char*) realloc(list_header_buf, 13 * sizeof(char));
			if (list_header_buf == NULL)
			{
				syslog(LOG_DEBUG, "Malloc for list_header_buf failed (location: 23).\n");
				return SMFIS_TEMPFAIL;
			}
			snprintf(list_header_buf, 13, "!Self SZPEM!");
			smfi_addheader(ctx, "Subject", list_header_buf);
			if (debug == 1)
				syslog(LOG_DEBUG, "List-Milter: changed subject to: %s.", list_header_buf);
		}

		free(list_header_buf);

		smfi_addheader(ctx, "X-List-Milter", "self-spam mail");
		if (debug == 1)
			syslog(LOG_DEBUG, "List-Milter: added header: self-spam mail.");
	}
	else if (priv->clamavness == 1)
	{
		if (priv->have_subject == 1)
		{
			len2 = (strlen(priv->subject) + 12);
			len = MIN(len2, 2047);
			if (len == 0)
			{
				syslog(LOG_DEBUG, "Realloc for list_header_buf: len = 0 (location: 25a).\n");
				return SMFIS_TEMPFAIL;
			}
			list_header_buf = (char*) realloc(list_header_buf, len * sizeof(char));
			if (list_header_buf == NULL)
			{
				syslog(LOG_DEBUG, "Malloc for list_header_buf failed (location: 25b).\n");
				return SMFIS_TEMPFAIL;
			}
			snprintf(list_header_buf, len, "!INFECTED! %s", priv->subject);
			smfi_chgheader(ctx, "Subject", 1, list_header_buf);
			if (debug == 1)
				syslog(LOG_DEBUG, "List-Milter: changed subject to: %s.", list_header_buf);
		}
		else
		{
			list_header_buf = (char*) realloc(list_header_buf, 11 * sizeof(char));
			if (list_header_buf == NULL)
			{
				syslog(LOG_DEBUG, "Malloc for list_header_buf failed (location: 26).\n");
				return SMFIS_TEMPFAIL;
			}
			snprintf(list_header_buf, 11, "!INFECTED!");
			smfi_addheader(ctx, "Subject", list_header_buf);
			if (debug == 1)
				syslog(LOG_DEBUG, "List-Milter: changed subject to: %s.", list_header_buf);
		}

		free(list_header_buf);

		smfi_addheader(ctx, "X-List-Milter", "infected mail");
		if (debug == 1)
			syslog(LOG_DEBUG, "List-Milter: added header: infected mail.");
	}
	else if (priv->local_mail == 1)
	{
		smfi_addheader(ctx, "X-List-Milter", "local mail");
		if (debug == 1)
			syslog(LOG_DEBUG, "List-Milter: added header: local mail.");
	}
	else
	{
		smfi_addheader(ctx, "X-List-Milter", "non-list mail");
		if (debug == 1)
			syslog(LOG_DEBUG, "List-Milter: added header: non-list mail.");
	}

	if (priv->have_subject == 1)
	{
		if (priv->dupl_subject == 1)
		{
			smfi_chgheader(ctx, "Subject", priv->subject_idx, NULL);
		}
	}


	if (debug == 1)
		syslog(LOG_DEBUG, "List-Milter: clue: %d.\n", priv->clue);

	return SMFIS_CONTINUE;
}

/*
 *   Log aborted connections
 */
sfsistat mlfi_abort(SMFICTX *ctx)
{
	struct mlfiPriv *priv = smfi_getpriv(ctx);

	syslog(LOG_DEBUG, "Don't blame me: other filter aborted the message:");
	syslog(LOG_DEBUG, "listness: %d, spammness: %d, clamavness: %d, local_mail: %d.", priv->listness, priv->spamness, priv->clamavness, priv->local_mail);

	return SMFIS_CONTINUE;
}

/*
 *   Free whatever I allocated
 */
sfsistat mlfi_close(SMFICTX *ctx)
{
	struct mlfiPriv *priv;

	priv = smfi_getpriv(ctx);

	if (priv != NULL)
	{
		priv->mailing_list = NULL;

		if (priv->subject != NULL)
			free(priv->subject);
		priv->subject = NULL;

/*		if (priv->mail_hostp != NULL)
			free(priv->mail_hostp);
		priv->mail_hostp = NULL;

		if (priv->mail_addrp != NULL)
			free(priv->mail_addrp);
		priv->mail_addrp = NULL;

		if (priv->rcpt_hostp != NULL)
			free(priv->rcpt_hostp);
		priv->rcpt_hostp = NULL; */

		if (priv->rcpt_addrp != NULL)
			free(priv->rcpt_addrp);
		priv->rcpt_addrp = NULL;

		if (priv->header_from != NULL)
			free(priv->header_from);
		priv->header_from = NULL;

		if (priv->header_to != NULL)
			free(priv->header_to);
		priv->header_to = NULL;

		if (priv->header_cc != NULL)
			free(priv->header_cc);
		priv->header_cc = NULL;

		if (priv->return_path != NULL)
			free(priv->return_path);
		priv->return_path = NULL;

		priv->listness = 0;
		priv->subjectness = 0;
		priv->clue = 0;

		priv->have_from = 0;
		priv->have_to = 0;
		priv->have_cc = 0;
		priv->have_subject = 0;
		priv->dupl_subject = 0;
		priv->subject_idx = 0;

		priv->local_mail = 0;
		priv->clamavness = 0;
		priv->spamness = 0;
		priv->returnpath = 0;

		smfi_setpriv(ctx, NULL);
		free(priv);
	}

	if (debug == 1)
		syslog(LOG_DEBUG, "Connection closed.");

	return SMFIS_CONTINUE;
}

struct smfiDesc smfilter =
{
	"list-milter",
	SMFI_VERSION,
	SMFIF_ADDHDRS | SMFIF_CHGHDRS,
	mlfi_connect,	/* xxfi_connect, */
	NULL,		/* xxfi_helo, */
	NULL,
/*	mlfi_envfrom, */	/* xxfi_envfrom, */
	mlfi_envrcpt,	/* xxfi_envrcpt, */
	mlfi_header,	/* xxfi_header, */
	NULL,		/* xxfi_eoh, */
	NULL,		/* xxfi_body, */
	mlfi_eom,	/* xxfi_eom, */
	mlfi_abort,	/* xxfi_abort, */
	mlfi_close,	/* xxfi_close */
	NULL,		/* xxfi_unknown, */
	NULL,		/* xxfi_data, */
	NULL,		/* xxfi_negotiate, */
};

void print_usage(char* prog_name)
{
	printf("Usage:\n");
	printf("%s -p socket-Addr -c conf-file [-h] [-d]\n\n", prog_name);
	printf("  -p\tListen to the inet/unix socket defined as Socket-Addr\n");
	printf("  -c\tConf-File name to use\n");
	printf("  -h\tHelp, a.k.a. this message\n");
	printf("  -d\tDebug mode, currently only more verbose output to syslog\n");
	printf("\n\n");
}

void parse_cmdline(int argc, char **argv)
{
	int i;
	int help_needed = 0;
	int expecting_socket = 0;
	int expecting_conffile = 0;

	for (i = 1; i < argc; i++)
	{
		if (expecting_socket == 1)
		{
			sock_addr = argv[i];
			expecting_socket = 0;
			continue;
		}
		if (expecting_conffile == 1)
		{
			conf_file = argv[i];
			expecting_conffile = 0;
			continue;
		}
		if (strcmp(argv[i], "-d") == 0)
		{
			debug = 1;
			continue;
		}
		if (strcmp(argv[i], "-h") == 0)
		{
			help_needed = 1;
			break;
		}
		if (strcmp(argv[i], "-p") == 0)
		{
			expecting_socket = 1;
			continue;
		}
		if (strcmp(argv[i], "-c") == 0)
		{
			expecting_conffile = 1;
			continue;
		}
	}
	if ((help_needed == 1)
		|| (expecting_socket == 1)
		|| (expecting_conffile == 1)
		|| (sock_addr == NULL))
	{
		print_usage(argv[0]);
		if (help_needed == 1)
			exit(EX_OK);
		else
			exit(EX_USAGE);
	}
}

void full_domainname()
{
/*	char *uts; -Wall unused variable */
/*	int i, j; -Wall unused variable */
	struct hostent *machine;
	size_t len = 1;

	host_name = (char *)malloc(len);
	while(gethostname(host_name, len) == -1 && errno == ENAMETOOLONG)
		host_name = (char *)realloc(host_name, ++len);
	if (debug == 1)
		printf("Hostname: %s\n", host_name);

	len = HOST_NAME_MAX;
	domain_name = (char *)malloc(len);
	while(getdomainname(domain_name, len) == -1 && errno == EINVAL)
		domain_name = (char *)realloc(domain_name, --len);
	if (debug == 1)
		printf("Domainname: %s\n", domain_name);

	machine = malloc(sizeof *machine);
	if (machine == NULL)
	{
		fprintf(stderr, "Malloc machine failed (location: 30).\n");
		exit(30);
	}
 	memset(machine, '\0', sizeof *machine);
	machine = gethostbyname(host_name);
	if (machine == NULL)
	{
		fprintf(stderr, "Gethostbyname failed (location: 31).\n");
		exit(31);
	}
	len = ((strlen(machine->h_name) + 1) * sizeof(char));
	if ((fqdn = (char*) malloc(len)) == NULL)
	{
		fprintf(stderr, "Malloc for fqdn failed (location: 32).\n");
		exit(32);
	}
	snprintf(fqdn, len, "%s", machine->h_name);
	if (debug == 1)
		printf("FQDN: %s, fqdn: %s\n\n", machine->h_name, fqdn);
}

void parse_confdefault()
{
	FILE *file_pointer;
	int i, k;
/*	int l; Wall: l is set but not used */
	char listname_buffer[HOST_NAME_MAX];

	file_pointer = fopen("/etc/mail/aliases", "r");
	if (file_pointer == NULL)
	{
		fprintf(stderr, "fopen(/etc/mail/aliases) failed (location: 33).\n");
		exit(33);
	}

	for (i = 0, listno = 1, k = 0; k != EOF; i++, listno++)
	{
		switch (k = fscanf(file_pointer, "owner-%63[^#@:\n]\n", listname_buffer))
		{
			case 1:
				list_array = realloc(list_array, listno * sizeof *list_array);
				if (list_array == NULL)
					fprintf(stderr, "Realloc for list_array failed (location: 34).\n");
				list_array[i] = malloc((strlen(listname_buffer) + 1) * sizeof(char));
				if (list_array[i] == NULL)
					fprintf(stderr, "Malloc for list_array[%d] failed (location: 35).\n", i);
				snprintf(list_array[i], (strlen(listname_buffer) + 1), "%s", listname_buffer);

				if (debug == 1)
					printf("list_array[%d]: %s\n", listno, list_array[i]);
				break;
			case 0:
/*				l = fscanf(file_pointer, "%*[^\n]\n"); Wall: l is set but not used  */
				fscanf(file_pointer, "%*[^\n]\n");
				i--;
				listno--;
				break;
			case EOF:
				printf("k = EOF: %d\n", k);
				i--;
				listno--;
				break;
		}
	}
	fclose(file_pointer);
	listno = i;
	if (debug == 1)
		printf("listno: %d\n", listno);
}

void parse_conffile()
{
	FILE *file_pointer;
	int i, k;
/*	int l; Wall: l is set but not used */
	char listname_buffer[HOST_NAME_MAX];
	size_t len;

	file_pointer = fopen(conf_file, "r");
	if (file_pointer == NULL)
	{
		fprintf(stderr, "fopen(%s) failed (location 35b)\n", conf_file);
		exit(35);
	}

	for (i = 0, listno = 1, k = 0; k != EOF; i++, listno++)
	{
		switch (k = fscanf(file_pointer, "%63[^#@\n]\n", listname_buffer))
		{
			case 1:
				list_array = realloc(list_array, listno * sizeof *list_array);
				if (list_array == NULL)
					fprintf(stderr, "Realloc for list_array failed (location: 36).\n");
				len = (strlen(listname_buffer) + 1);
				list_array[i] = malloc(len * sizeof(char));
				if (list_array[i] == NULL)
					fprintf(stderr, "Malloc for list_array[%d] failed (location: 37).\n", i);
				snprintf(list_array[i], len, "%s", listname_buffer);

				if (debug == 1)
					printf("list_array[%d]: %s, listno: %d\n", i, list_array[i], listno);
				break;
			case 0:
/*				l = fscanf(file_pointer, "%*[^\n]\n"); Wall: l is set but not used  */
				fscanf(file_pointer, "%*[^\n]\n");
				i--;
				listno--;
				break;
			case EOF:
				if (debug == 1)
					printf("k = EOF: %d\n", k);
				i--;
				listno--;
				break;
		}
	}
	fclose(file_pointer);
	listno = i;
	if (debug == 1)
		printf("listno: %d\n", listno);
}

/*
 *   Standard way to daemonize, fork twice to make sure the final child is not a
 *   session leader. Then you have to close all open file descriptors to detach
 *   from the console, also change directory somewhere safe and reset umask.
 *   TODO: Error checking
 */
int daemonize()
{
	int r, s, t, u;

	r = fork();
	if (r < 0)		/* failed */
		return -1;
	else if (r != 0)	/* parent */
		exit(EX_OK);
	else			/* child */
	{
		setsid();
		r = fork();
		if (r < 0)		/* failed */
			return -1;
		else if (r != 0)	/* parent */
			exit(EX_OK);
		else			/* child */
		{
			close(0); close(1); close(2);	/* close old fds */
			open("/dev/null", O_RDWR);	/* new fd 0 */
			s = dup(0); t = dup(0);		/* and fds 1 and 2 */
			if (s < 0 || t  < 0)
			{
				syslog(LOG_DEBUG, "Duplicating file descriptor failed while demonizing (location: 38).\n");
			}

			umask(0);			/* clear umask */
			u = chdir("/var/tmp");		/* change to /var/tmp */
			if (u != 0)
			{
				syslog(LOG_DEBUG, "Failed to change to /var/tmp while demonizing (location: 39).\n");
			}
			return 0;
		}
	}
}

/*
 *   A milter's main() has to do only 3 things:
 *   1) Set the listening socket with smfi_setcon()
 *   2) Register the callbacks table with smfi_register()
 *   3) Pass control to smfi_main
 *
 *   Simple, isn't it?
 */
int main(int argc, char **argv) 
{
	int rc;

	parse_cmdline(argc, argv);

	if (conf_file == NULL)
		parse_confdefault();
	else
		parse_conffile();

	full_domainname();

	openlog(argv[0], LOG_PID | LOG_NDELAY, LOG_MAIL);

	smfi_setconn(sock_addr);

	rc = smfi_register(smfilter);
	if (rc == MI_FAILURE)
	{
		fprintf(stderr,"smfi_register() failed (location: 40).\n");
		exit(EX_UNAVAILABLE);
	}

	if (daemonize() != 0)
	{
		fprintf(stderr, "daemonization failed (location: 41).\n");
		exit(EX_UNAVAILABLE);
	}

	rc = smfi_main();
	closelog();
	return rc;
}

/*
cdecoder.c - c source to a base64 decoding algorithm implementation

This is part of the libb64 project, and has been placed in the public domain.
For details, see http://sourceforge.net/projects/libb64
*/

int base64_decode_value(char value_in)
{
	static const char decoding[] = {62,-1,-1,-1,63,52,53,54,55,56,57,58,59,60,61,-1,-1,-1,-2,-1,-1,-1,0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,-1,-1,-1,-1,-1,-1,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51};
	static const char decoding_size = sizeof(decoding);
	value_in -= 43;
	if (value_in < 0 || value_in > decoding_size) return -1;
	return decoding[(int)value_in];
}

void base64_init_decodestate(base64_decodestate* state_in)
{
	state_in->step = step_a;
	state_in->plainchar = 0;
}

int base64_decode_block(const char* code_in, const int length_in, char* plaintext_out, base64_decodestate* state_in)
{
	const char* codechar = code_in;
	char* plainchar = plaintext_out;
	char fragment;
	
	*plainchar = state_in->plainchar;
	
	switch (state_in->step)
	{
		while (1)
		{
	case step_a:
			do {
				if (codechar == code_in+length_in)
				{
					state_in->step = step_a;
					state_in->plainchar = *plainchar;
					return plainchar - plaintext_out;
				}
				fragment = (char)base64_decode_value(*codechar++);
			} while (fragment < 0);
			*plainchar    = (fragment & 0x03f) << 2;
	case step_b:
			do {
				if (codechar == code_in+length_in)
				{
					state_in->step = step_b;
					state_in->plainchar = *plainchar;
					return plainchar - plaintext_out;
				}
				fragment = (char)base64_decode_value(*codechar++);
			} while (fragment < 0);
			*plainchar++ |= (fragment & 0x030) >> 4;
			*plainchar    = (fragment & 0x00f) << 4;
	case step_c:
			do {
				if (codechar == code_in+length_in)
				{
					state_in->step = step_c;
					state_in->plainchar = *plainchar;
					return plainchar - plaintext_out;
				}
				fragment = (char)base64_decode_value(*codechar++);
			} while (fragment < 0);
			*plainchar++ |= (fragment & 0x03c) >> 2;
			*plainchar    = (fragment & 0x003) << 6;
	case step_d:
			do {
				if (codechar == code_in+length_in)
				{
					state_in->step = step_d;
					state_in->plainchar = *plainchar;
					return plainchar - plaintext_out;
				}
				fragment = (char)base64_decode_value(*codechar++);
			} while (fragment < 0);
			*plainchar++   |= (fragment & 0x03f);
		}
	}
	/* control should not reach here */
	return plainchar - plaintext_out;
}


/* taken from: http://stackoverflow.com/questions/779875/what-is-the-function-to-replace-string-in-c */
char *replace(char const * const original, char const * const pattern, char const * const replacement)
{
	size_t const replen = strlen(replacement);
	size_t const patlen = strlen(pattern);
	size_t const orilen = strlen(original);

	size_t patcnt = 0;
	const char *oriptr;
	const char *patloc;

	/* find how many times the pattern occurs in the original string */
	for (oriptr = original; (patloc = strstr(oriptr, pattern)); oriptr = patloc + patlen)
	{
		patcnt++;
	}

	{
		/* allocate memory for the new string */
		size_t const retlen = orilen + patcnt * (replen - patlen);
		char * const returned = (char *) malloc( sizeof(char) * (retlen + 1) );

		if (returned != NULL)
		{
			/* copy the original string, */
			/* replacing all the instances of the pattern */
			char * retptr = returned;
			for (oriptr = original; (patloc = strstr(oriptr, pattern)); oriptr = patloc + patlen)
			{
				size_t const skplen = patloc - oriptr;
				/* copy the section until the occurence of the pattern */
				strncpy(retptr, oriptr, skplen);
				retptr += skplen;
				/* copy the replacement */
				strncpy(retptr, replacement, replen);
				retptr += replen;
			}
			/* copy the rest of the string. */
			strcpy(retptr, oriptr);
		}
		return returned;
	}
}
